"""CLI handling for `routemaster`."""
import logging

import yaml
import click

from routemaster.app import App
from routemaster.cron import CronThread
from routemaster.config import ConfigError, load_config
from routemaster.server import server
from routemaster.middleware import wrap_application
from routemaster.validation import ValidationError, validate_config
from routemaster.gunicorn_application import GunicornWSGIApplication

logger = logging.getLogger(__name__)


@click.group()
@click.option(
    '-c',
    '--config-file',
    help="Path to the service config file.",
    type=click.File(encoding='utf-8'),
    required=True,
)
@click.pass_context
def main(ctx, config_file):
    """Shared entrypoint configuration."""
    logging.getLogger('schedule').setLevel(logging.CRITICAL)

    try:
        config = load_config(yaml.load(config_file))
    except ConfigError:
        logger.exception("Configuration Error")
        click.get_current_context().exit(1)

    ctx.obj = App(config)
    _validate_config(ctx.obj)


@main.command()
@click.pass_context
def validate(ctx):
    """
    Entrypoint for validation of configuration files.

    Validation is done by the main handler in order to cover all code paths,
    so this function is a stub so that `serve` does not have to be called.
    """
    pass


@main.command()
@click.option(
    '-b',
    '--bind',
    help="Bind address and port.",
    type=str,
    default='[::]:2017',
)
@click.option(
    '--debug/--no-debug',
    help="Enable debugging mode.",
    default=False,
)
@click.option(
    '--workers',
    help="Number of gunicorn workers to run.",
    type=int,
    default=1,
)
@click.pass_context
def serve(ctx, bind, debug, workers):  # pragma: no cover
    """Entrypoint for serving the Routemaster HTTP service."""
    app = ctx.obj

    server.config.app = app
    if debug:
        server.config['DEBUG'] = True

    cron_thread = CronThread(app)
    cron_thread.start()

    wrapped_server = wrap_application(app, server)

    def post_fork():
        app.initialise()
        app.logger.init_flask(server)

    try:
        instance = GunicornWSGIApplication(
            wrapped_server,
            bind=bind,
            debug=debug,
            workers=workers,
            post_fork=post_fork,
        )
        instance.run()
    finally:
        cron_thread.stop()


def _validate_config(app: App):
    try:
        validate_config(app, app.config)
    except ValidationError as e:
        msg = f"Validation Error: {e}"
        logger.exception(msg)
        click.get_current_context().exit(1)



"""Initialisation for database connections."""
from sqlalchemy import create_engine
from sqlalchemy.engine import Engine

from routemaster.config import DatabaseConfig


def initialise_db(config: DatabaseConfig) -> Engine:
    """Initialise a database given the connection string."""
    engine = create_engine(config.connstr)
    return engine



"""
Main entry point for `routemaster` command-line utility.

This is to enable `python -m routemaster` if that is needed for any reason,
normal use should be to use the `routemaster` command-line tool directly.
"""

from routemaster.cli import main

main()



"""Processing for gate states."""
from sqlalchemy import func

from routemaster.db import Label, History
from routemaster.app import App
from routemaster.config import Gate, State, StateMachine
from routemaster.state_machine.types import LabelRef
from routemaster.state_machine.utils import (
    choose_next_state,
    context_for_label,
    get_state_machine,
    get_label_metadata,
    get_current_history,
)
from routemaster.state_machine.exceptions import DeletedLabel


def process_gate(
    *,
    app: App,
    state: State,
    state_machine: StateMachine,
    label: LabelRef,
) -> bool:
    """
    Process a label in a gate, continuing if necessary.

    Assumes that `gate` is the current state of the label, and that the label
    has been locked.

    Returns whether the label progressed in the state machine, for which `True`
    implies further progression should be attempted.
    """
    if not isinstance(state, Gate):  # pragma: no branch
        raise ValueError(  # pragma: no cover
            f"process_gate called with {state.name} which is not an Gate",
        )

    gate = state

    state_machine = get_state_machine(app, label)
    metadata, deleted = get_label_metadata(app, label, state_machine)
    if deleted:
        raise DeletedLabel(label)

    history_entry = get_current_history(app, label)

    context = context_for_label(
        label,
        metadata,
        state_machine,
        gate,
        history_entry,
        app.logger,
    )
    can_exit = gate.exit_condition.run(context)

    if not can_exit:
        return False

    destination = choose_next_state(state_machine, gate, context)

    app.session.add(History(
        label_state_machine=state_machine.name,
        label_name=label.name,
        created=func.now(),
        old_state=gate.name,
        new_state=destination.name,
    ))

    app.session.query(Label).filter_by(
        name=label.name,
        state_machine=label.state_machine,
    ).update({
        'metadata_triggers_processed': True,
    })

    return True



"""Core API endpoints for routemaster service."""

import sqlalchemy
import pkg_resources
from flask import Flask, abort, jsonify, request

from routemaster import state_machine
from routemaster.state_machine import (
    LabelRef,
    UnknownLabel,
    LabelAlreadyExists,
    UnknownStateMachine,
)

server = Flask('routemaster')
server.config['PROPAGATE_EXCEPTIONS'] = True


@server.route('/', methods=['GET'])
def status():
    """
    Status check endpoint.

    Returns:
    - 200 Ok:                  if upstream services are up and the application
                               appears ready to serve requests.
    - 503 Service Unavailable: if there is any detected reason why the service
                               might not be able to serve requests.
    """
    try:
        version = pkg_resources.working_set.by_key['routemaster'].version
    except KeyError:  # pragma: no cover
        version = 'development'

    try:
        server.config.app.session.query(sqlalchemy.literal(1)).one()
        return jsonify({
            'status': 'ok',
            'state-machines': '/state-machines',
            'version': version,
        })
    except Exception:
        return jsonify({
            'status': 'error',
            'message': 'Cannot connect to database',
            'version': version,
        }), 503


@server.route('/state-machines', methods=['GET'])
def get_state_machines():
    """
    List the state machines known to this server.

    Successful return codes return a list of dictionaries containing at least
    the name of each state machine.
    """
    return jsonify({
        'state-machines': [
            {
                'name': x.name,
                'labels': f'/state-machines/{x.name}/labels',
            }
            for x in server.config.app.config.state_machines.values()
        ],
    })


@server.route(
    '/state-machines/<state_machine_name>/labels',
    methods=['GET'],
)
def get_labels(state_machine_name):
    """
    List the labels in a state machine.

    Returns:
    - 200 Ok: if the state machine exists.
    - 404 Not Found: if the state machine does not exist.

    Successful return codes return a list of dictionaries containing at least
    the name of each label.
    """
    app = server.config.app

    try:
        state_machine_instance = app.config.state_machines[state_machine_name]
    except KeyError:
        msg = f"State machine '{state_machine_name}' does not exist"
        abort(404, msg)

    labels = state_machine.list_labels(app, state_machine_instance)
    return jsonify({
        'labels': [{'name': x.name} for x in labels],
        'create': f'/state-machines/{state_machine_name}/labels/:name',
    })


@server.route(
    '/state-machines/<state_machine_name>/labels/<label_name>',
    methods=['GET'],
)
def get_label(state_machine_name, label_name):
    """
    Get a label within a given state machine.

    Returns:
    - 200 Ok: if the label is exists.
    - 404 Not Found: if the state machine or label does not exist.
    - 410 Gone: if the label once existed but has since been deleted.

    Successful return codes return the full metadata for the label.
    """
    app = server.config.app
    label = LabelRef(label_name, state_machine_name)

    try:
        metadata = state_machine.get_label_metadata(app, label)
    except UnknownLabel as e:
        abort(
            410 if e.deleted else 404,
            f"Label {label.name} in state machine '{label.state_machine}' "
            f"does not exist.",
        )
    except UnknownStateMachine:
        abort(404, f"State machine '{label.state_machine}' does not exist.")

    state = state_machine.get_label_state(app, label)
    return jsonify(metadata=metadata, state=state.name)


@server.route(
    '/state-machines/<state_machine_name>/labels/<label_name>',
    methods=['POST'],
)
def create_label(state_machine_name, label_name):
    """
    Create a label with a given metadata, and start it in the state machine.

    Returns:
    - 201 Created: if the label is successfully created and started.
    - 409 Conflict: if the label already exists in the state machine.
    - 404 Not Found: if the state machine does not exist.
    - 400 Bad Request: if the request body is not a valid metadata.

    Successful return codes return the full created metadata for the label.
    """
    app = server.config.app
    label = LabelRef(label_name, state_machine_name)
    data = request.get_json()

    try:
        initial_metadata = data['metadata']
    except KeyError:
        abort(400, "No metadata given")

    try:
        initial_state_name = \
            app.config.state_machines[state_machine_name].states[0].name
        metadata = state_machine.create_label(app, label, initial_metadata)
        return jsonify(metadata=metadata, state=initial_state_name), 201
    except LookupError:
        msg = f"State machine '{state_machine_name}' does not exist"
        abort(404, msg)
    except LabelAlreadyExists:
        msg = f"Label {label_name} already exists in '{state_machine_name}'"
        abort(409, msg)


@server.route(
    '/state-machines/<state_machine_name>/labels/<label_name>', # noqa
    methods=['PATCH'],
)
def update_label(state_machine_name, label_name):
    """
    Update a label in a state machine.

    Triggering progression if necessary according to the state machine
    configuration. Updates are _merged_ with existing metadata.

    Returns:
    - 200 Ok: if the label is successfully updated.
    - 400 Bad Request: if the request body is not a valid metadata.
    - 404 Not Found: if the state machine or label does not exist.
    - 410 Gone: if the label once existed but has since been deleted.

    Successful return codes return the full new metadata for a label.
    """
    app = server.config.app
    label = LabelRef(label_name, state_machine_name)

    try:
        patch_metadata = request.get_json()['metadata']
    except KeyError:
        abort(400, "No new metadata")

    try:
        new_metadata = state_machine.update_metadata_for_label(
            app,
            label,
            patch_metadata,
        )
        state = state_machine.get_label_state(app, label)
        return jsonify(metadata=new_metadata, state=state.name)
    except UnknownStateMachine:
        msg = f"State machine '{state_machine_name}' does not exist"
        abort(404, msg)
    except UnknownLabel as e:
        abort(
            410 if e.deleted else 404,
            f"Label {label_name} does not exist in state machine "
            f"'{state_machine_name}'.",
        )


@server.route(
    '/state-machines/<state_machine_name>/labels/<label_name>',
    methods=['DELETE'],
)
def delete_label(state_machine_name, label_name):
    """
    Delete a label in a state machine.

    Marks a label as deleted, but does not remove it from the database.
    Deleted labels cannot be updated and will not move state.

    Returns:
    - 204 No content: if the label is successfully deleted (or did not exist).
    - 404 Not Found: if the state machine does not exist.
    """
    app = server.config.app
    label = LabelRef(label_name, state_machine_name)

    try:
        state_machine.delete_label(app, label)
    except UnknownStateMachine:
        msg = f"State machine '{state_machine_name}' does not exist"
        abort(404, msg)

    return '', 204


@server.route('/check-loggers', methods=['GET'])
def check_loggers():
    """
    Logging check endpoint.

    This endpoint errors so as to verify that loggers are working.
    """
    raise Exception("Test exception")



"""Entry point for debugging purposes."""

import sys

from routemaster.exit_conditions.parser import parse
from routemaster.exit_conditions.peephole import peephole_optimise
from routemaster.exit_conditions.exceptions import ParseError
from routemaster.exit_conditions.error_display import (
    format_parse_error_message,
)

source = sys.stdin.read()
try:
    for instruction, *args in peephole_optimise(parse(source)):
        sys.stdout.write(
            f"{instruction.value} {', '.join(repr(x) for x in args)}",
        )
except ParseError as e:
    sys.stdout.write(format_parse_error_message(
        source=source,
        error=e,
    ))



"""Utilities for state machine execution."""

import datetime
import functools
import contextlib
from typing import Any, Dict, List, Tuple, Optional, Sequence, Collection

import dateutil.tz
from sqlalchemy import func

from routemaster.db import Label, History
from routemaster.app import App
from routemaster.feeds import feeds_for_state_machine
from routemaster.config import Gate, State, StateMachine, ContextNextStates
from routemaster.context import Context
from routemaster.logging import BaseLogger
from routemaster.state_machine.types import LabelRef, Metadata
from routemaster.state_machine.exceptions import (
    UnknownLabel,
    UnknownStateMachine,
)


def get_state_machine(app: App, label: LabelRef) -> StateMachine:
    """Finds the state machine instance by name in the app config."""
    try:
        return app.config.state_machines[label.state_machine]
    except KeyError:
        raise UnknownStateMachine(label.state_machine)


def choose_next_state(
    state_machine: StateMachine,
    current_state: State,
    context: Context,
) -> State:
    """Assuming a transition is valid, choose a next state."""
    next_state_name = current_state.next_states.next_state_for_label(context)
    return state_machine.get_state(next_state_name)


def get_label_metadata(
    app: App,
    label: LabelRef,
    state_machine: StateMachine,
) -> Tuple[Dict[str, Any], bool]:
    """Get the metadata and whether the label has been deleted."""
    return app.session.query(Label.metadata, Label.deleted).filter_by(
        name=label.name,
        state_machine=state_machine.name,
    ).first()


def get_current_state(
    app: App,
    label: LabelRef,
    state_machine: StateMachine,
) -> Optional[State]:
    """Get the current state of a label, based on its last history entry."""
    history_entry = get_current_history(app, label)
    if history_entry.new_state is None:
        # label has been deleted
        return None
    return state_machine.get_state(history_entry.new_state)


def get_current_history(app: App, label: LabelRef) -> History:
    """Get a label's last history entry."""
    history_entry = app.session.query(History).filter_by(
        label_name=label.name,
        label_state_machine=label.state_machine,
    ).order_by(
        # Our model type stubs define the `id` attribute as `int`, yet
        # sqlalchemy actually allows the attribute to be used for ordering like
        # this; ignore the type check here specifically rather than complicate
        # our type definitions.
        History.id.desc(),  # type: ignore
    ).first()

    if history_entry is None:
        raise UnknownLabel(label)

    return history_entry


def needs_gate_evaluation_for_metadata_change(
    app: App,
    state_machine: StateMachine,
    label: LabelRef,
    update: Metadata,
) -> Tuple[bool, State]:
    """
    Given a change to the metadata, should the gate evaluation be triggered.
    """

    current_state = get_current_state(app, label, state_machine)

    if current_state is None:
        raise ValueError(
            f"Cannot determine gate evaluation for deleted label {label} "
            "(deleted labels have no current state)",
        )

    if not isinstance(current_state, Gate):
        # Label is not a gate state so there's no trigger to resolve.
        return False, current_state

    if any(
        trigger.should_trigger_for_update(update)
        for trigger in current_state.metadata_triggers
    ):
        return True, current_state

    return False, current_state


def lock_label(app: App, label: LabelRef) -> Label:
    """Lock a label in the current transaction."""
    row = app.session.query(Label).filter_by(
        name=label.name,
        state_machine=label.state_machine,
    ).with_for_update().first()

    if row is None:
        raise UnknownLabel(label)

    return row


def labels_in_state(
    app: App,
    state_machine: StateMachine,
    state: State,
) -> List[str]:
    """Util to get all the labels in an action state that need retrying."""
    return _labels_in_state(app, state_machine, state, True)


def labels_in_state_with_metadata(
    app: App,
    state_machine: StateMachine,
    state: State,
    path: Sequence[str],
    values: Collection[str],
) -> List[str]:
    """
    Util to get all the labels in a given state with some metadata value.

    The metadata lookup happens at the given path, allowing for any of the
    posible values given.
    """
    if not values:
        raise ValueError("Must specify at least one possible value")

    metadata_lookup = Label.metadata
    for part in path:
        metadata_lookup = metadata_lookup[part]  # type: ignore

    return _labels_in_state(
        app,
        state_machine,
        state,
        metadata_lookup.astext.in_(values),  # type: ignore
    )


def labels_needing_metadata_update_retry_in_gate(
    app: App,
    state_machine: StateMachine,
    state: State,
) -> List[str]:
    """Util to get all the labels in an action state that need retrying."""
    if not isinstance(state, Gate):  # pragma: no branch
        raise ValueError(  # pragma: no cover
            f"labels_needing_metadata_update_retry_in_gate called with "
            f"{state.name} which is not an Gate",
        )

    return _labels_in_state(
        app,
        state_machine,
        state,
        ~Label.metadata_triggers_processed,
    )


def _labels_in_state(
    app: App,
    state_machine: StateMachine,
    state: State,
    filter_: Any,
) -> List[str]:
    """Util to get all the labels in an action state that need retrying."""

    states_by_rank = app.session.query(
        History.label_name,
        History.new_state,
        func.row_number().over(
            # Our model type stubs define the `id` attribute as `int`, yet
            # sqlalchemy actually allows the attribute to be used for ordering
            # like this; ignore the type check here specifically rather than
            # complicate our type definitions.
            order_by=History.id.desc(),  # type: ignore
            partition_by=History.label_name,
        ).label('rank'),
    ).filter_by(
        label_state_machine=state_machine.name,
    ).subquery()

    ranked_transitions = app.session.query(
        states_by_rank.c.label_name,
    ).filter(
        states_by_rank.c.rank == 1,
        states_by_rank.c.new_state == state.name,
    ).join(Label).filter(
        filter_,
    )

    return [x for x, in ranked_transitions]


def context_for_label(
    label: LabelRef,
    metadata: Metadata,
    state_machine: StateMachine,
    state: State,
    history_entry: Any,
    logger: BaseLogger,
) -> Context:
    """Util to build the context for a label in a state."""
    feeds = feeds_for_state_machine(state_machine)

    accessed_variables: List[str] = []
    if isinstance(state, Gate):
        accessed_variables.extend(state.exit_condition.accessed_variables())
    if isinstance(state.next_states, ContextNextStates):
        accessed_variables.append(state.next_states.path)

    @contextlib.contextmanager
    def feed_logging_context(feed_url):
        with logger.process_feed(state_machine, state, feed_url):
            yield functools.partial(
                logger.feed_response,
                state_machine,
                state,
                feed_url,
            )

    return Context(
        label=label.name,
        metadata=metadata,
        now=datetime.datetime.now(dateutil.tz.tzutc()),
        feeds=feeds,
        accessed_variables=accessed_variables,
        current_history_entry=history_entry,
        feed_logging_context=feed_logging_context,
    )



"""Logging plugin subsystem."""

from routemaster.logging.base import BaseLogger
from routemaster.logging.plugins import (
    PluginConfigurationException,
    register_loggers,
)
from routemaster.logging.split_logger import SplitLogger
from routemaster.logging.python_logger import PythonLogger

__all__ = (
    'BaseLogger',
    'SplitLogger',
    'PythonLogger',
    'register_loggers',
    'PluginConfigurationException',
)



"""Package setup."""

import version
from setuptools import setup, find_packages

with open('README.md', 'r', encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='routemaster',
    version=version.get_version(),
    url='https://github.com/thread/routemaster',
    description="State machines as a service.",
    long_description=long_description,
    long_description_content_type="text/markdown",

    author="Thread",
    author_email="tech@thread.com",

    keywords=(
    ),
    license='MIT',

    zip_safe=False,

    packages=find_packages(),
    include_package_data=True,

    classifiers=(
        'Development Status :: 2 - Pre-Alpha',
        'Environment :: Console',
        'Natural Language :: English',
        'Operating System :: POSIX',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.6',
        'Topic :: Office/Business',
    ),

    install_requires=(
        'click',
        'pyyaml',
        'jsonschema >=2.6, <3',
        'flask',
        'psycopg2-binary',
        'sqlalchemy',
        'python-dateutil',
        'alembic >=0.9.6',
        'gunicorn >=19.7',
        'schedule',
        'freezegun',
        'requests',
        'networkx',
        'dataclasses',
        'typing-extensions>=3.7.4',
    ),

    setup_requires=(
        'pytest-runner',
    ),

    tests_require=(
        'pytest',
        'networkx',
        'tox',
        'pytest-cov',
        'pytest-pythonpath',
    ),

    entry_points={
        'console_scripts': (
            'routemaster = routemaster.cli:main',
        ),
    },
)



"""Validation of state machines."""
import collections

import networkx
from sqlalchemy import func

from routemaster.db import History
from routemaster.app import App
from routemaster.config import Config, StateMachine


class ValidationError(Exception):
    """Class for errors raised to indicate invalid configuration."""
    pass


def validate_config(app: App, config: Config):
    """Validate that a given config satisfies invariants."""
    for state_machine in config.state_machines.values():
        _validate_state_machine(app, state_machine)


def _validate_state_machine(app: App, state_machine: StateMachine):
    """Validate that a given state machine is internally consistent."""
    with app.new_session():
        _validate_route_start_to_end(state_machine)
        _validate_all_states_exist(state_machine)
        _validate_no_labels_in_nonexistent_states(state_machine, app)
        _validate_unique_state_names(state_machine)


def _build_graph(state_machine: StateMachine) -> networkx.Graph:
    graph = networkx.Graph()
    for state in state_machine.states:
        graph.add_node(state.name)
        for destination_name in state.next_states.all_destinations():
            graph.add_edge(state.name, destination_name)
    return graph


def _validate_route_start_to_end(state_machine):
    graph = _build_graph(state_machine)
    if not networkx.is_connected(graph):
        raise ValidationError("Graph is not fully connected")


def _validate_unique_state_names(state_machine):
    state_name_counts = collections.Counter([
        x.name for x in state_machine.states
    ])

    invalid_states = [x for x, y in state_name_counts.items() if y > 1]

    if invalid_states:
        raise ValidationError(
            f"States {invalid_states!r} are not unique in "
            f"{state_machine.name}",
        )


def _validate_all_states_exist(state_machine):
    state_names = set(x.name for x in state_machine.states)
    for state in state_machine.states:
        for destination_name in state.next_states.all_destinations():
            if destination_name not in state_names:
                raise ValidationError(f"{destination_name} does not exist")


def _validate_no_labels_in_nonexistent_states(state_machine, app):
    states = [x.name for x in state_machine.states]

    states_by_rank = app.session.query(
        History.label_name,
        History.new_state,
        func.row_number().over(
            order_by=History.id.desc(),
            partition_by=History.label_name,
        ).label('rank'),
    ).filter_by(
        label_state_machine=state_machine.name,
    ).subquery()

    invalid_labels_and_states = app.session.query(
        states_by_rank.c.label_name,
        states_by_rank.c.new_state,
    ).filter(
        states_by_rank.c.rank == 1,
        ~(
            states_by_rank.c.new_state.in_(states) |
            states_by_rank.c.new_state.is_(None)
        ),
    ).all()

    if invalid_labels_and_states:
        raise ValidationError(
            f"{len(invalid_labels_and_states)} nodes in states that no "
            f"longer exist",
        )



"""Shared utilities."""
import contextlib
from typing import Any, Dict, Callable, Iterable, Sequence

StartResponse = Callable[
    [
        str,
        Dict[str, str],
        Any,
    ],
    None,
]

WSGIEnvironment = Dict[str, Any]

WSGICallable = Callable[
    [
        WSGIEnvironment,
        StartResponse,
    ],
    Iterable[bytes],
]


def dict_merge(d1, d2):
    """
    Recursively merge two dicts to create a new dict.

    Does not modify either dict. The second dict's values will take priority
    over the first's.
    """
    new = dict(d1)
    for k, v in d2.items():
        if (
            k in new and
            isinstance(new[k], dict) and
            isinstance(d2[k], dict)
        ):
            new[k] = dict_merge(new[k], d2[k])
        else:
            new[k] = d2[k]
    return new


def get_path(path: Sequence[str], d: Dict[str, Any]) -> Any:
    """Get the path from a structure of nested dicts."""
    if not len(path):
        # Empty path returns the whole dict, i.e. no _filter_ on the dict
        return d

    component, rest = path[0], path[1:]
    if rest:
        return get_path(rest, d.get(component, {}))
    return d.get(component)


@contextlib.contextmanager
def suppress_exceptions(logger):
    """Catch all exceptions and log to a provided logger."""
    try:
        yield
    except Exception:
        logger.exception("Error suppressed")


def template_url(
    url_template: str,
    state_machine_name: str,
    label: str,
) -> str:
    """
    Templates a URL for an external service.

    Adds the label and state machine to a url that contains placeholders in the
    format `<label>` or '<state_machine>'.
    """
    return url_template.replace(
        '<label>',
        label,
    ).replace(
        '<state_machine>',
        state_machine_name,
    )



"""Global test setup and fixtures."""

import os
import re
import json
import socket
import datetime
import functools
import contextlib
import subprocess
from typing import Any, Dict
from unittest import mock

import pytest
import httpretty
import dateutil.tz
import pkg_resources
from sqlalchemy import create_engine
from werkzeug.test import Client
from sqlalchemy.orm import sessionmaker
from werkzeug.wrappers import BaseResponse

from routemaster import state_machine
from routemaster.db import Label, History, metadata
from routemaster.app import App
from routemaster.utils import dict_merge
from routemaster.config import (
    Gate,
    Action,
    Config,
    Webhook,
    FeedConfig,
    NoNextStates,
    StateMachine,
    DatabaseConfig,
    OnEntryTrigger,
    MetadataTrigger,
    ConstantNextState,
    ContextNextStates,
    LoggingPluginConfig,
    ContextNextStatesOption,
)
from routemaster.server import server
from routemaster.context import Context
from routemaster.logging import BaseLogger, SplitLogger, register_loggers
from routemaster.webhooks import (
    WebhookResult,
    webhook_runner_for_state_machine,
)
from routemaster.middleware import wrap_application
from routemaster.state_machine import LabelRef
from routemaster.exit_conditions import ExitConditionProgram

TEST_DATABASE_CONFIG = DatabaseConfig(
    host=os.environ.get('PG_HOST', 'localhost'),
    port=int(os.environ.get('PG_PORT', 5432)),
    name=os.environ.get('PG_DB', 'routemaster_test'),
    username=os.environ.get('PG_USER', ''),
    password=os.environ.get('PG_PASS', ''),
)

TEST_STATE_MACHINES = {
    'test_machine': StateMachine(
        name='test_machine',
        feeds=[
            FeedConfig(name='tests', url='http://localhost/tests'),
        ],
        webhooks=[
            Webhook(
                match=re.compile('//(.+\\.)?example\\.com'),
                headers={
                    'x-api-key': 'Rahfew7eed1ierae0moa2sho3ieB1et3ohhum0Ei',
                },
            ),
        ],
        states=[
            Gate(
                name='start',
                triggers=[
                    MetadataTrigger(metadata_path='should_progress'),
                    OnEntryTrigger(),
                ],
                next_states=ContextNextStates(
                    path='feeds.tests.should_do_alternate_action',
                    destinations=[
                        ContextNextStatesOption(
                            state='perform_action',
                            value=False,
                        ),
                        ContextNextStatesOption(
                            state='perform_alternate_action',
                            value=True,
                        ),
                    ],
                    default='perform_action',
                ),
                exit_condition=ExitConditionProgram(
                    'metadata.should_progress = true',
                ),
            ),
            Action(
                name='perform_action',
                webhook='http://localhost/hook/<state_machine>/<label>',
                next_states=ConstantNextState(state='end'),
            ),
            Action(
                name='perform_alternate_action',
                webhook='http://localhost/hook',
                next_states=ContextNextStates(
                    path='feeds.tests.should_loop',
                    destinations=[
                        ContextNextStatesOption(
                            state='end',
                            value=False,
                        ),
                        ContextNextStatesOption(
                            state='start',
                            value=True,
                        ),
                    ],
                    default='end',
                ),
            ),
            Gate(
                name='end',
                triggers=[],
                next_states=NoNextStates(),
                exit_condition=ExitConditionProgram('false'),
            ),
        ],
    ),
    # This state machine is used for exercising race conditions in tests and is
    # purposefully not realistic.
    'test_machine_2': StateMachine(
        name='test_machine_2',
        feeds=[],
        webhooks=[],
        states=[
            Gate(
                name='gate_1',
                triggers=[
                    OnEntryTrigger(),
                    MetadataTrigger(metadata_path='should_progress'),
                ],
                next_states=ConstantNextState('gate_2'),
                exit_condition=ExitConditionProgram(
                    'metadata.should_progress = true',
                ),
            ),
            Gate(
                name='gate_2',
                triggers=[
                    MetadataTrigger(metadata_path='should_progress'),
                ],
                next_states=ConstantNextState('end'),
                exit_condition=ExitConditionProgram(
                    'metadata.should_progress = true',
                ),
            ),
            Gate(
                name='end',
                triggers=[],
                next_states=NoNextStates(),
                exit_condition=ExitConditionProgram('false'),
            ),
        ],
    ),
    # This state machine is used for testing the possibility of infinite loops.
    'test_infinite_machine': StateMachine(
        name='test_infinite_machine',
        feeds=[],
        webhooks=[],
        states=[
            Gate(
                name='gate_1',
                triggers=[
                    OnEntryTrigger(),
                    MetadataTrigger(metadata_path='should_progress'),
                ],
                next_states=ConstantNextState('gate_2'),
                exit_condition=ExitConditionProgram(
                    'metadata.should_progress = true',
                ),
            ),
            Gate(
                name='gate_2',
                triggers=[OnEntryTrigger()],
                next_states=ConstantNextState('gate_3'),
                exit_condition=ExitConditionProgram('true'),
            ),
            Gate(
                name='gate_3',
                triggers=[OnEntryTrigger()],
                next_states=ConstantNextState('gate_2'),
                exit_condition=ExitConditionProgram('true'),
            ),
        ],
    ),
    'test_machine_timing': StateMachine(
        name='test_machine_timing',
        feeds=[],
        webhooks=[],
        states=[
            Gate(
                name='start',
                triggers=[
                    OnEntryTrigger(),
                ],
                next_states=ConstantNextState('end'),
                exit_condition=ExitConditionProgram(
                    '1d12h has passed since history.entered_state',
                ),
            ),
            Gate(
                name='end',
                triggers=[],
                next_states=NoNextStates(),
                exit_condition=ExitConditionProgram('false'),
            ),
        ],
    ),
}

TEST_ENGINE = create_engine(TEST_DATABASE_CONFIG.connstr)


class TestApp(App):
    """
    Mocked App subclass which overloads the `db` property.

    This has two implications:

    1. We use a shared engine in all the tests rather than connecting many
       times (a speed improvement),
    2. We can set a flag on access to `.db` so that we needn't bother with
       resetting the database if nothing has actually been changed.
    """
    def __init__(self, config):
        self.config = config
        self.session_used = False
        self.logger = SplitLogger(config, loggers=register_loggers(config))
        self._session = None
        self._needs_rollback = False
        self._current_session = None
        self._sessionmaker = sessionmaker(bind=TEST_ENGINE)
        self._webhook_runners = {
            x: webhook_runner_for_state_machine(y)
            for x, y in self.config.state_machines.items()
        }

    @property
    def session(self):
        """Start if necessary and return a shared session."""
        self.session_used = True
        return super().session


class TestClientResponse(BaseResponse):
    """Test client response format."""

    @property
    def json(self):
        """Util property for json responses."""
        return json.loads(self.data)


def get_test_app(**kwargs):
    """Instantiate an app with testing parameters."""
    return TestApp(Config(
        state_machines=kwargs.get('state_machines', TEST_STATE_MACHINES),
        database=kwargs.get('database', TEST_DATABASE_CONFIG),
        logging_plugins=kwargs.get('logging_plugins', [
            LoggingPluginConfig(
                dotted_path='routemaster.logging:PythonLogger',
                kwargs={'log_level': 'DEBUG'},
            ),
        ]),
    ))


@pytest.fixture()
def client(custom_app=None):
    """Create a werkzeug test client."""
    _app = get_test_app() if custom_app is None else custom_app
    server.config.app = _app
    _app.logger.init_flask(server)
    return Client(wrap_application(_app, server), TestClientResponse)


@pytest.fixture()
def app(**kwargs):
    """Create an `App` config object for testing."""
    return get_test_app(**kwargs)


@pytest.fixture()
def custom_app():
    """Return the test app generator so that we can pass in custom config."""
    return get_test_app


@pytest.fixture()
def app_env():
    """
    Create a dict of environment variables.

    Mirrors the testing environment for use in subprocesses.
    """
    return {
        'DB_HOST': TEST_DATABASE_CONFIG.host,
        'DB_PORT': str(TEST_DATABASE_CONFIG.port),
        'DB_NAME': TEST_DATABASE_CONFIG.name,
        'DB_USER': TEST_DATABASE_CONFIG.username,
        'DB_PASS': TEST_DATABASE_CONFIG.password,
    }


@pytest.fixture(autouse=True, scope='session')
def database_creation(request):
    """Wrap test session in creating and destroying all required tables."""
    metadata.drop_all(bind=TEST_ENGINE)
    metadata.create_all(bind=TEST_ENGINE)
    yield


@pytest.yield_fixture(autouse=True)
def database_clear(app):
    """Truncate all tables after each test."""
    yield
    if app.session_used:
        with app.new_session():
            for table in metadata.tables:
                app.session.execute(
                    f'truncate table {table} cascade',
                    {},
                )


@pytest.fixture()
def create_label(app, mock_test_feed):
    """Create a label in the database."""

    def _create(
        name: str,
        state_machine_name: str,
        metadata: Dict[str, Any],
    ) -> LabelRef:
        with mock_test_feed(), app.new_session():
            state_machine.create_label(
                app,
                LabelRef(name, state_machine_name),
                metadata,
            )
            return LabelRef(name, state_machine_name)

    return _create


@pytest.fixture()
def delete_label(app):
    """
    Mark a label in the database as deleted.
    """

    def _delete(name: str, state_machine_name: str) -> None:
        with app.new_session():
            state_machine.delete_label(
                app,
                LabelRef(name, state_machine_name),
            )

    return _delete


@pytest.fixture()
def create_deleted_label(create_label, delete_label):
    """
    Create a label in the database and then delete it.
    """

    def _create_and_delete(name: str, state_machine_name: str) -> LabelRef:
        create_label(name, state_machine_name, {})
        delete_label(name, state_machine_name)
        return LabelRef(name, state_machine_name)

    return _create_and_delete


@pytest.fixture()
def mock_webhook():
    """Mock the test config's webhook call."""
    @contextlib.contextmanager
    def _mock(result=WebhookResult.SUCCESS):
        runner = mock.Mock(return_value=result)
        with mock.patch(
            'routemaster.app.App.get_webhook_runner',
            return_value=runner,
        ):
            yield runner
    return _mock


@pytest.fixture()
def mock_test_feed():
    """Mock out the test feed."""
    @contextlib.contextmanager
    def _mock(data={'should_do_alternate_action': False}):
        httpretty.enable()
        httpretty.register_uri(
            httpretty.GET,
            'http://localhost/tests',
            body=json.dumps(data),
            content_type='application/json',
        )

        try:
            yield
        finally:
            httpretty.disable()
            httpretty.reset()

    return _mock


@pytest.fixture()
def assert_history(app):
    """Assert that the database history matches what is expected."""
    def _assert(entries):
        with app.new_session():
            history_entries = [
                (x.old_state, x.new_state)
                for x in app.session.query(
                    History,
                ).order_by(
                    History.id,
                )
            ]

            assert history_entries == entries
    return _assert


@pytest.fixture()
def set_metadata(app):
    """Directly set the metadata for a label in the database."""
    def _inner(label, update):
        with app.new_session():
            db_label = app.session.query(Label).filter_by(
                name=label.name,
                state_machine=label.state_machine,
            ).first()

            db_label.metadata = dict_merge(db_label.metadata, update)
            db_label.metadata_triggers_processed = True

            return db_label.metadata
    return _inner


@pytest.fixture()
def make_context(app):
    """Factory for Contexts that provides sane defaults for testing."""
    def _inner(**kwargs):
        logger = BaseLogger(app.config)
        state_machine = app.config.state_machines['test_machine']
        state = state_machine.states[0]

        @contextlib.contextmanager
        def feed_logging_context(feed_url):
            with logger.process_feed(state_machine, state, feed_url):
                yield functools.partial(
                    logger.feed_response,
                    state_machine,
                    state,
                    feed_url,
                )

        return Context(
            label=kwargs['label'],
            metadata=kwargs.get('metadata', {}),
            now=kwargs.get('now', datetime.datetime.now(dateutil.tz.tzutc())),
            feeds=kwargs.get('feeds', {}),
            accessed_variables=kwargs.get('accessed_variables', []),
            current_history_entry=kwargs.get('current_history_entry'),
            feed_logging_context=kwargs.get(
                'feed_logging_context',
                feed_logging_context,
            ),
        )
    return _inner


@pytest.fixture()
def version():
    """Return the package version."""
    try:
        return pkg_resources.working_set.by_key['routemaster'].version
    except KeyError:  # pragma: no cover
        return 'development'


@pytest.fixture()
def current_state(app):
    """Get the current state of a label."""
    def _inner(label):
        with app.new_session():
            return app.session.query(
                History.new_state,
            ).filter_by(
                label_name=label.name,
                label_state_machine=label.state_machine,
            ).order_by(
                History.id.desc(),
            ).limit(1).scalar()
    return _inner


@pytest.fixture()
def unused_tcp_port():
    """Returns an unused TCP port, inspired by pytest-asyncio."""
    with contextlib.closing(socket.socket()) as sock:
        sock.bind(('127.0.0.1', 0))
        return sock.getsockname()[1]


@pytest.fixture()
def routemaster_serve_subprocess(unused_tcp_port):
    """
    Fixture to spawn a routemaster server as a subprocess.

    Yields the process reference, and the port that it can be accessed on.
    """

    @contextlib.contextmanager
    def _inner(*, wait_for_output=None):
        env = os.environ.copy()
        env.update({
            'DB_HOST': os.environ.get('PG_HOST', 'localhost'),
            'DB_PORT': os.environ.get('PG_PORT', '5432'),
            'DB_NAME': os.environ.get('PG_DB', 'routemaster_test'),
            'DB_USER': os.environ.get('PG_USER', ''),
            'DB_PASS': os.environ.get('PG_PASS', ''),
        })

        try:
            proc = subprocess.Popen(
                [
                    'routemaster',
                    '--config-file=example.yaml',
                    'serve',
                    '--bind',
                    f'127.0.0.1:{unused_tcp_port}',
                ],
                env=env,
                stdin=subprocess.PIPE,
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT,
            )

            if wait_for_output is not None:
                all_output = b''
                while True:
                    assert proc.poll() is None, all_output.decode('utf-8')

                    out_line = proc.stdout.readline()
                    if wait_for_output in out_line:
                        break

                    all_output += out_line

            yield proc, unused_tcp_port
        finally:
            proc.terminate()

    return _inner
